/****************************************************************************
 *
 *      $Id: main.c,v 1.6 2003/08/06 22:52:04 benjl Exp $
 *
 *      Description: Example startup file for the SOS project.
 *
 *      Author:      Ben Leslie
 *
 ****************************************************************************/

#include <types.h>
#include <l4.h>
#include <l4e.h>
#include <dite.h>
#include <stddef.h>
#include <stdlib.h>
#include <assert.h>
#include <serial.h>

#include "frames.h"
#include "pager.h"
#include "hardware.h"

#define HEAP_SIZE (1 * 1024 * 1024) /* 1 MB heap */

extern void irq_handler(L4_ThreadId_t tid, L4_MsgTag_t tag, L4_Msg_t *msg);

struct tulip *tulip;

#define verbose 0
#define dprintf(v, args...)			\
    if (verbose > (v))				\
	l4e_printf (args)

static void syscall_loop (void);
static void init_thread(void);

/* Set aside some memory for a stack for the
   first user task */

#define STACK_SIZE 0x1000
static uint64_t user_stack[STACK_SIZE];
static uint64_t init_stack[STACK_SIZE];

/*
  Main entry point - called by crt.
*/

int
main (void)
{
	L4_Word_t high, low;
	L4_ThreadId_t setup_thread_tid;

	/* Initialise the L4 Environment library */
	l4e_init();

	/* Announce ourself to the world */
	l4e_printf("SOS Starting. My thread id is  %lx\n\n", L4_Myself());
	
	/* Find information about available memory */
	l4e_find_memory(&low, &high);

	l4e_printf("Available memory from %lx to %lx - %dMB\n", 
	       low, high, (high - low) / 1024 / 1024);

	/* Initialise the malloc heap */
	__malloc_init((void*) low, (void*) (low + HEAP_SIZE));

	/* Initialise the frame table */
	frame_init((low + HEAP_SIZE), high);

	/* 
	   Spawn the setup thread which completes the rest of the 
	   initialisation, leaving this thread free to act as a pager
	   and interrupt handler.
	*/
	setup_thread_tid = l4e_get_new_tid();
	l4e_start_thread(setup_thread_tid, init_thread, &init_stack[STACK_SIZE-1]);

	/* Enter the syscall loop */
	syscall_loop();

	return 0;
}

/*
  Init thread - This function starts up our device drivers and runs the
  first user program.
*/
static void
init_thread(void)
{
	struct Dite_Dhdr *dhdr;
	struct Dite_Phdr *phdr;
	int i, taskid = 1;

	/* Initialise the u4600 hardware */
	hardware_init();

	/* Get access to the dit header */
	dhdr = dite_gethdr();
	assert(dhdr != NULL);

	/* Loop through the dit and start up user programs */
	for (i=1; i < dhdr->d_phnum; i++) {
		L4_ThreadId_t newtid;
		phdr = dite_getentry(dhdr, i);
		assert(phdr != NULL);

		l4e_printf("Found: %d %s\n", i, phdr->p_name);

		/* Start a new task with this program */
		newtid = l4e_task_2_tid(taskid++);
		l4e_printf("Created tid: %lx\n", newtid);
		l4e_task_new(newtid, L4_Pager(), 
			     (void*)(L4_Word_t)phdr->p_entry, 
			     (void*) &user_stack[0x1000]);
	}

	/* Thread finished - sleep forever */
	L4_Sleep(L4_Never);
}


/* Some IPC labels defined in the L4 documentation */
#define L4_INTERRUPT  0xfffffffffff
#define L4_PAGEFAULT  0xffffffffffe
#define L4_SIGMA0_REQ 0xffffffffffa

/*
  Syscall loop.

  This implements a very simple, single threaded functions for 
  recieving any IPCs and dispatching to the correct subsystem.

  It currently handles pagefaults, interrupts and special sigma0
  requests.
*/

static void 
syscall_loop (void)
{
	L4_MsgTag_t tag;
	L4_ThreadId_t tid;
	L4_Msg_t msg;
	int send;

	/* Setup which messages we will recieve */
	L4_Accept (L4_UntypedWordsAcceptor);
	
	/* Wait for the first message */
	tag = L4_Wait (&tid);
	
	for(;;) {
		/* At this point we have recieved an IPC */

		L4_MsgStore (tag, &msg); /* Get the tag */
		send = 1; /* In most cases we will want to send a reply */
	
		dprintf (1, "syscall: got msg from %p, (0x%lx)\n", 
			 (void *) tid.raw, (long) L4_Label (tag),
			 (void *) L4_MsgWord (&msg, 0));
		
		/*
		 * Dispatch IPC according to protocol.
		 */
		switch (L4_Label(tag) >> 4) {
		case L4_PAGEFAULT:
			/* A pagefault occured. Dispatch to the pager */
			if (pager(tid, tag, &msg) != 0)
				send = 0; /* Don't want to reply if pager
					     call failed */
			break;
		case L4_INTERRUPT:
			/* An interrupt occured. Dispatch 
			   to interrupt handler */
			irq_handler(tid, tag, &msg);
			break;
		case L4_SIGMA0_REQ:
			/* A thread requested something from Sigma0.
			   This current implementation always forwards 
			   to Sigma0. If you care about security, you might
			   want to check that this only comes from a kernel
			   thread. (Such as the init_thread.) */

			/* We want to receive fpages from Sigma0 */
			L4_Accept(L4_MapGrantItems(L4_CompleteAddressSpace));
			
			/* Make the call */
			tag = L4_Call(L4_Pager());

			/* Check the result */
			if (L4_IpcFailed(tag)) {
				L4_Word_t ec = L4_ErrorCode();
				l4e_printf("l4_request_page: IPC error\n");
				l4e_print_error(ec);
			}

			/* Reset out accept mask */
			L4_Accept (L4_UntypedWordsAcceptor);
			break;
		default:
			dprintf (0, "syscall: unknown syscall request from %p, "
				 "(%p, %p, %p)\n",
				 (void *) tid.raw, (void *) tag.raw,
				 (void *) L4_MsgWord (&msg, 0), 
				 (void *) L4_MsgWord(&msg, 1));
			/* Unknown system call, so we don't want to reply
			   to this thread */
			send = 0;
		}
		
		if (send == 0) {
			/* Nothing to send, so we just wait */
			tag = L4_Wait (&tid);
		} else {
			/* Reply to caller and wait for next IPC */
			tag = L4_ReplyWait (tid, &tid);
		
			if (L4_IpcFailed(tag)) {
				L4_Word_t ec = L4_ErrorCode();
				l4e_printf("l4_request_page: IPC error\n");
				l4e_print_error(ec);
			}
		}
	}
}
